@@ -1,9 +1,9 @@
+import contextlib
 import io
 import os
 import re
 import sys
 
-import jsonschema
 import numpy as np
 import pytest
 import yaml
@@ -11,82 +11,155 @@ from numpy import ma
 from numpy.testing import assert_array_equal
 
 import asdf
-from asdf import util
-from asdf._tests import _helpers as helpers
-from asdf._tests.objects import CustomTestType
-from asdf.exceptions import AsdfDeprecationWarning
+from asdf.exceptions import ValidationError
+from asdf.extension import Converter, Extension, TagDefinition
 from asdf.tags.core import ndarray
-
-from . import data as test_data
-
-TEST_DATA_PATH = helpers.get_test_data_path("", module=test_data)
+from asdf.testing import helpers
 
 
 # These custom types and the custom extension are here purely for the purpose
 # of testing NDArray objects and making sure that they can be validated as part
 # of a nested hierarchy, and not just top-level objects.
-class CustomNdim(CustomTestType):
-    name = "ndim"
-    organization = "nowhere.org"
-    standard = "custom"
-    version = "1.0.0"
+class CustomData:
+    def __init__(self, value):
+        self.value = value
 
 
-class CustomDatatype(CustomTestType):
-    name = "datatype"
-    organization = "nowhere.org"
-    standard = "custom"
-    version = "1.0.0"
+class CustomNDim:
+    def __init__(self, value):
+        self.value = value
 
 
-class CustomExtension:
-    @property
-    def types(self):
-        return [CustomNdim, CustomDatatype]
+class CustomNDimConverter(Converter):
+    tags = ["tag:nowhere.org:custom/ndim-1.0.0"]
+    types = [CustomNDim]
 
-    @property
-    def tag_mapping(self):
-        return [("tag:nowhere.org:custom", "http://nowhere.org/schemas/custom{tag_suffix}")]
+    def to_yaml_tree(self, obj, tag, ctx):
+        return obj.value
 
-    @property
-    def url_mapping(self):
-        return [("http://nowhere.org/schemas/custom/", util.filepath_to_url(TEST_DATA_PATH) + "/{url_suffix}.yaml")]
+    def from_yaml_tree(self, node, tag, ctx):
+        return CustomNDim(node)
 
 
-def test_sharing(tmpdir):
+class CustomDataConverter(Converter):
+    tags = ["tag:nowhere.org:custom/datatype-1.0.0"]
+    types = [CustomData]
+
+    def to_yaml_tree(self, obj, tag, ctx):
+        return obj.value
+
+    def from_yaml_tree(self, node, tag, ctx):
+        return CustomData(node)
+
+
+class CustomExtension(Extension):
+    tags = [
+        TagDefinition(
+            tag_uri="tag:nowhere.org:custom/datatype-1.0.0",
+            schema_uris=["http://nowhere.org/schemas/custom/datatype-1.0.0"],
+        ),
+        TagDefinition(
+            tag_uri="tag:nowhere.org:custom/ndim-1.0.0",
+            schema_uris=["http://nowhere.org/schemas/custom/ndim-1.0.0"],
+        ),
+    ]
+    extension_uri = "asdf://nowhere.org/extensions/custom-1.0.0"
+    converters = [CustomDataConverter(), CustomNDimConverter()]
+
+
+@contextlib.contextmanager
+def with_custom_extension():
+    with asdf.config_context() as cfg:
+        cfg.add_extension(CustomExtension())
+        cfg.add_resource_mapping(
+            {
+                "http://nowhere.org/schemas/custom/datatype-1.0.0": """%YAML 1.1
+---
+$schema: "http://stsci.edu/schemas/asdf/asdf-schema-1.0.0"
+id: "http://nowhere.org/schemas/custom/datatype-1.0.0"
+type: object
+properties:
+  a:
+    datatype: float32
+
+  b:
+    datatype: float32
+    exact_datatype: true
+
+  c:
+    datatype:
+      - name: a
+        datatype: int16
+      - name: b
+        datatype: ['ascii', 16]
+
+  d:
+    datatype:
+      - name: a
+        datatype: int16
+      - name: b
+        datatype: ['ascii', 16]
+    exact_datatype: true""",
+                "http://nowhere.org/schemas/custom/ndim-1.0.0": """%YAML 1.1
+---
+$schema: "http://stsci.edu/schemas/asdf/asdf-schema-1.0.0"
+id: "http://nowhere.org/schemas/custom/ndim-1.0.0"
+type: object
+properties:
+  a:
+    ndim: 2
+
+  b:
+    max_ndim: 2""",
+            }
+        )
+        yield
+
+
+@contextlib.contextmanager
+def roundtrip(af, raw=False):
+    if not isinstance(af, asdf.AsdfFile):
+        af = asdf.AsdfFile(af)
+    b = io.BytesIO()
+    af.write_to(b)
+    b.seek(0)
+    if raw:
+        bs = b.read()
+        if asdf.constants.BLOCK_MAGIC in bs:
+            bs, *_ = bs.split(asdf.constants.BLOCK_MAGIC)
+        yield bs
+    else:
+        with asdf.open(b) as af:
+            yield af
+
+
+def test_sharing():
     x = np.arange(0, 10, dtype=float)
     tree = {"science_data": x, "subset": x[3:-3], "skipping": x[::2]}
 
-    def check_asdf(asdf):
-        tree = asdf.tree
-
+    with roundtrip(tree) as af:
+        tree = af.tree
         assert_array_equal(tree["science_data"], x)
         assert_array_equal(tree["subset"], x[3:-3])
         assert_array_equal(tree["skipping"], x[::2])
 
         assert tree["science_data"].ctypes.data == tree["skipping"].ctypes.data
 
-        assert len(list(asdf._blocks.internal_blocks)) == 1
-        assert next(asdf._blocks.internal_blocks)._size == 80
+        assert len(af._blocks.blocks) == 1
+        assert af._blocks.blocks[0].header["data_size"] == 80
 
-        if "w" in asdf._mode:
-            tree["science_data"][0] = 42
-            assert tree["skipping"][0] == 42
-
-    def check_raw_yaml(content):
-        assert b"!core/ndarray" in content
-
-    helpers.assert_roundtrip_tree(tree, tmpdir, asdf_check_func=check_asdf, raw_yaml_check_func=check_raw_yaml)
+        tree["science_data"][0] = 42
+        assert tree["skipping"][0] == 42
 
 
-def test_byteorder(tmpdir):
+def test_byteorder(tmp_path):
     tree = {
         "bigendian": np.arange(0, 10, dtype=">f8"),
         "little": np.arange(0, 10, dtype="<f8"),
     }
 
-    def check_asdf(asdf):
-        my_tree = asdf.tree
+    with roundtrip(tree) as af:
+        my_tree = af.tree
         for endian in ("bigendian", "little"):
             assert my_tree[endian].dtype == tree[endian].dtype
 
@@ -97,14 +170,8 @@ def test_byteorder(tmpdir):
             assert my_tree["bigendian"].dtype.byteorder == "="
             assert my_tree["little"].dtype.byteorder == "<"
 
-    def check_raw_yaml(content):
-        assert b"byteorder: little" in content
-        assert b"byteorder: big" in content
 
-    helpers.assert_roundtrip_tree(tree, tmpdir, asdf_check_func=check_asdf, raw_yaml_check_func=check_raw_yaml)
-
-
-def test_all_dtypes(tmpdir):
+def test_all_dtypes(tmp_path):
     tree = {}
     for byteorder in (">", "<"):
         for dtype in ndarray._datatype_names.values():
@@ -118,7 +185,11 @@ def test_all_dtypes(tmpdir):
 
             tree[byteorder + dtype] = arr
 
-    helpers.assert_roundtrip_tree(tree, tmpdir)
+    with roundtrip(tree) as af:
+        for k in tree:
+            pre = tree[k]
+            post = af[k]
+            assert_array_equal(pre, post)
 
 
 def test_dont_load_data():
@@ -131,18 +202,15 @@ def test_dont_load_data():
 
     buff.seek(0)
     with asdf.open(buff) as ff:
-        with pytest.warns(AsdfDeprecationWarning, match="AsdfFile.run_hook is deprecated"):
-            ff.run_hook("reserve_blocks")
-
         # repr and str shouldn't load data
         str(ff.tree["science_data"])
         repr(ff.tree)
 
-        for block in ff._blocks.internal_blocks:
-            assert block._data is None
+        for block in ff._blocks.blocks:
+            assert callable(block._data)
 
 
-def test_table_inline(tmpdir):
+def test_table_inline(tmp_path):
     table = np.array(
         [(0, 1, (2, 3)), (4, 5, (6, 7))],
         dtype=[("MINE", np.int8), ("", np.float64), ("arr", ">i4", (2,))],
@@ -150,36 +218,73 @@ def test_table_inline(tmpdir):
 
     tree = {"table_data": table}
 
-    def check_raw_yaml(content):
-        tree = yaml.safe_load(re.sub(rb"!core/\S+", b"", content))
-
-        assert tree["table_data"] == {
-            "datatype": [
-                {"datatype": "int8", "name": "MINE"},
-                {"datatype": "float64", "name": "f1"},
-                {"datatype": "int32", "name": "arr", "shape": [2]},
-            ],
-            "data": [[0, 1.0, [2, 3]], [4, 5.0, [6, 7]]],
-            "shape": [2],
-        }
-
     with asdf.config_context() as config:
         config.array_inline_threshold = 100
-        helpers.assert_roundtrip_tree(tree, tmpdir, raw_yaml_check_func=check_raw_yaml)
+
+        with roundtrip(tree) as af:
+            assert table.dtype.names == af["table_data"].dtype.names
+            for n in table.dtype.names:
+                assert_array_equal(table[n], af["table_data"][n])
+
+        with roundtrip(tree, raw=True) as content:
+            tree = yaml.safe_load(re.sub(rb"!core/\S+", b"", content))
+
+            assert tree["table_data"] == {
+                "datatype": [
+                    {"datatype": "int8", "name": "MINE"},
+                    {"datatype": "float64", "name": "f1"},
+                    {"datatype": "int32", "name": "arr", "shape": [2]},
+                ],
+                "data": [[0, 1.0, [2, 3]], [4, 5.0, [6, 7]]],
+                "shape": [2],
+            }
 
 
-def test_array_inline_threshold_recursive(tmpdir):
-    models = pytest.importorskip("astropy.modeling.models")
+def test_array_inline_threshold_recursive(tmp_path):
+    """
+    Test that setting the inline threshold works for objects
+    that contain (and when serialized produce a ndarray)
+    """
 
-    aff = models.AffineTransformation2D(matrix=[[1, 2], [3, 4]])
-    tree = {"test": aff}
+    class NDArrayContainer:
+        def __init__(self, array):
+            self._array = array
 
-    def check_asdf(asdf):
-        assert len(list(asdf._blocks.internal_blocks)) == 0
+        @property
+        def array(self):
+            return np.array(self._array)
+
+    class NDArrayContainerConverter:
+        tags = ["http://somewhere.org/tags/foo-1.0.0"]
+        types = [NDArrayContainer]
+
+        def to_yaml_tree(self, obj, tag, ctx):
+            return {"array": obj.array}
+
+        def from_yaml_tree(self, node, tag, ctx):
+            return NDArrayContainer(node["array"])
+
+    class NDArrayContainerExtension:
+        tags = NDArrayContainerConverter.tags
+        converters = [NDArrayContainerConverter()]
+        extension_uri = "http://somewhere.org/extensions/foo-1.0.0"
+
+    container = NDArrayContainer([[1, 2], [3, 4]])
+    tree = {"test": container}
 
     with asdf.config_context() as config:
+        config.add_extension(NDArrayContainerExtension())
         config.array_inline_threshold = 100
-        helpers.assert_roundtrip_tree(tree, tmpdir, asdf_check_func=check_asdf)
+        # we can no longer use _helpers.assert_roundtrip_tree here because
+        # the model no longer has a CustomType which results in equality testing
+        # using == which will fail
+        # this test appears to be designed to test the inline threshold so we can
+        # just look at the number of blocks
+        fn = str(tmp_path / "test.asdf")
+        af = asdf.AsdfFile(tree)
+        af.write_to(fn)
+        with asdf.open(fn) as af:
+            assert len(list(af._blocks.blocks)) == 0
 
 
 def test_copy_inline():
@@ -196,12 +301,15 @@ x0: !core/ndarray-1.0.0
         f.write_to(io.BytesIO())
 
 
-def test_table(tmpdir):
+def test_table(tmp_path):
     table = np.array([(0, 1, (2, 3)), (4, 5, (6, 7))], dtype=[("MINE", np.int8), ("", "<f8"), ("arr", ">i4", (2,))])
 
     tree = {"table_data": table}
 
-    def check_raw_yaml(content):
+    with roundtrip(tree) as af:
+        assert_array_equal(af["table_data"], table)
+
+    with roundtrip(tree, raw=True) as content:
         tree = yaml.safe_load(re.sub(rb"!core/\S+", b"", content))
 
         assert tree["table_data"] == {
@@ -215,10 +323,8 @@ def test_table(tmpdir):
             "byteorder": "big",
         }
 
-    helpers.assert_roundtrip_tree(tree, tmpdir, raw_yaml_check_func=check_raw_yaml)
 
-
-def test_table_nested_fields(tmpdir):
+def test_table_nested_fields(tmp_path):
     table = np.array(
         [(0, (1, 2)), (4, (5, 6)), (7, (8, 9))],
         dtype=[("A", "<i8"), ("B", [("C", "<i8"), ("D", "<i8")])],
@@ -226,7 +332,10 @@ def test_table_nested_fields(tmpdir):
 
     tree = {"table_data": table}
 
-    def check_raw_yaml(content):
+    with roundtrip(tree) as af:
+        assert_array_equal(af["table_data"], table)
+
+    with roundtrip(tree, raw=True) as content:
         tree = yaml.safe_load(re.sub(rb"!core/\S+", b"", content))
 
         assert tree["table_data"] == {
@@ -246,8 +355,6 @@ def test_table_nested_fields(tmpdir):
             "byteorder": "big",
         }
 
-    helpers.assert_roundtrip_tree(tree, tmpdir, raw_yaml_check_func=check_raw_yaml)
-
 
 def test_inline():
     x = np.arange(0, 10, dtype=float)
@@ -256,13 +363,14 @@ def test_inline():
     buff = io.BytesIO()
 
     ff = asdf.AsdfFile(tree)
-    ff._blocks.set_array_storage(ff._blocks[tree["science_data"]], "inline")
+    ff.set_array_storage(x, "inline")
     ff.write_to(buff)
 
     buff.seek(0)
     with asdf.open(buff, mode="rw") as ff:
-        helpers.assert_tree_match(tree, ff.tree)
-        assert len(list(ff._blocks.internal_blocks)) == 0
+        for k in tree:
+            assert_array_equal(tree[k], ff.tree[k])
+        assert len(list(ff._blocks.blocks)) == 0
         buff = io.BytesIO()
         ff.write_to(buff)
 
@@ -277,38 +385,34 @@ def test_inline_bare():
         assert_array_equal(ff.tree["arr"], [[1, 2, 3, 4], [5, 6, 7, 8]])
 
 
-def test_mask_roundtrip(tmpdir):
+def test_mask_roundtrip(tmp_path):
     x = np.arange(0, 10, dtype=float)
     m = ma.array(x, mask=x > 5)
     tree = {"masked_array": m, "unmasked_array": x}
 
-    def check_asdf(asdf):
-        tree = asdf.tree
+    with roundtrip(tree) as af:
+        tree = af.tree
 
         m = tree["masked_array"]
 
         assert np.all(m.mask[6:])
-        assert len(asdf._blocks) == 2
-
-    helpers.assert_roundtrip_tree(tree, tmpdir, asdf_check_func=check_asdf)
+        assert len(af._blocks.blocks) == 2
 
 
-def test_len_roundtrip(tmpdir):
+def test_len_roundtrip(tmp_path):
     sequence = np.arange(0, 10, dtype=int)
     tree = {"sequence": sequence}
 
-    def check_len(asdf):
-        s = asdf.tree["sequence"]
+    with roundtrip(tree) as af:
+        s = af.tree["sequence"]
         assert len(s) == 10
 
-    helpers.assert_roundtrip_tree(tree, tmpdir, asdf_check_func=check_len)
-
 
 def test_mask_arbitrary():
     content = """
-    arr: !core/ndarray-1.0.0
-      data: [[1, 2, 3, 1234], [5, 6, 7, 8]]
-      mask: 1234
+arr: !core/ndarray-1.0.0
+  data: [[1, 2, 3, 1234], [5, 6, 7, 8]]
+  mask: 1234
     """
 
     buff = helpers.yaml_to_asdf(content)
@@ -318,9 +422,9 @@ def test_mask_arbitrary():
 
 def test_mask_nan():
     content = """
-    arr: !core/ndarray-1.0.0
-      data: [[1, 2, 3, .NaN], [5, 6, 7, 8]]
-      mask: .NaN
+arr: !core/ndarray-1.0.0
+  data: [[1, 2, 3, .NaN], [5, 6, 7, 8]]
+  mask: .NaN
     """
 
     buff = helpers.yaml_to_asdf(content)
@@ -328,19 +432,23 @@ def test_mask_nan():
         assert_array_equal(ff.tree["arr"].mask, [[False, False, False, True], [False, False, False, False]])
 
 
-def test_string(tmpdir):
+def test_string(tmp_path):
     tree = {
         "ascii": np.array([b"foo", b"bar", b"baz"]),
-        "unicode": np.array(["სამეცნიერო", "данные", "வடிவம்"]),  # noqa: RUF001
+        "unicode": np.array(["სამეცნიერო", "данные", "வடிவம்"]),
     }
 
-    helpers.assert_roundtrip_tree(tree, tmpdir)
+    with roundtrip(tree) as af:
+        for k in tree:
+            assert_array_equal(tree[k], af[k])
 
 
-def test_string_table(tmpdir):
+def test_string_table(tmp_path):
     tree = {"table": np.array([(b"foo", "სამეცნიერო", "42", "53.0")])}
 
-    helpers.assert_roundtrip_tree(tree, tmpdir)
+    with roundtrip(tree) as af:
+        for k in tree:
+            assert_array_equal(tree[k], af[k])
 
 
 def test_inline_string():
@@ -353,12 +461,12 @@ def test_inline_string():
 
 def test_inline_structured():
     content = """
-    arr: !core/ndarray-1.0.0
-        datatype: [['ascii', 4], uint16, uint16, ['ascii', 4]]
-        data: [[M110, 110, 205, And],
-               [ M31,  31, 224, And],
-               [ M32,  32, 221, And],
-               [M103, 103, 581, Cas]]"""
+arr: !core/ndarray-1.0.0
+    datatype: [['ascii', 4], uint16, uint16, ['ascii', 4]]
+    data: [[M110, 110, 205, And],
+           [ M31,  31, 224, And],
+           [ M32,  32, 221, And],
+           [M103, 103, 581, Cas]]"""
 
     buff = helpers.yaml_to_asdf(content)
 
@@ -393,7 +501,7 @@ def test_simple_table():
     ff.write_to(io.BytesIO())
 
 
-def test_unicode_to_list(tmpdir):
+def test_unicode_to_list(tmp_path):
     arr = np.array(["", "𐀠"], dtype="<U")
     tree = {"unicode": arr}
 
@@ -404,12 +512,12 @@ def test_unicode_to_list(tmpdir):
     fd.seek(0)
 
     with asdf.open(fd) as ff:
-        ff.resolve_and_inline()
-        ff.write_to(io.BytesIO())
+        ff.resolve_references()
+        ff.write_to(io.BytesIO(), all_array_storage="inline")
 
 
-def test_inline_masked_array(tmpdir):
-    testfile = os.path.join(str(tmpdir), "masked.asdf")
+def test_inline_masked_array(tmp_path):
+    testfile = os.path.join(str(tmp_path), "masked.asdf")
 
     tree = {"test": ma.array([1, 2, 3], mask=[0, 1, 0])}
 
@@ -418,17 +526,17 @@ def test_inline_masked_array(tmpdir):
     f.write_to(testfile)
 
     with asdf.open(testfile) as f2:
-        assert len(list(f2._blocks.internal_blocks)) == 0
+        assert len(list(f2._blocks.blocks)) == 0
         assert_array_equal(f.tree["test"], f2.tree["test"])
 
     with open(testfile, "rb") as fd:
         assert b"null" in fd.read()
 
 
-def test_masked_array_stay_open_bug(tmpdir):
+def test_masked_array_stay_open_bug(tmp_path):
     psutil = pytest.importorskip("psutil")
 
-    tmppath = os.path.join(str(tmpdir), "masked.asdf")
+    tmppath = os.path.join(str(tmp_path), "masked.asdf")
 
     tree = {"test": np.ma.array([1, 2, 3], mask=[False, True, False])}
 
@@ -445,7 +553,7 @@ def test_masked_array_stay_open_bug(tmpdir):
     assert len(p.open_files()) <= len(orig_open)
 
 
-def test_memmap_stay_open_bug(tmpdir):
+def test_memmap_stay_open_bug(tmp_path):
     """
     Regression test for issue #1239
     memmapped arrays only closed at garbage collection when asdf.open given an open file
@@ -458,7 +566,7 @@ def test_memmap_stay_open_bug(tmpdir):
     """
     psutil = pytest.importorskip("psutil")
 
-    tmppath = os.path.join(str(tmpdir), "arr.asdf")
+    tmppath = os.path.join(str(tmp_path), "arr.asdf")
 
     tree = {"test": np.array([1, 2, 3])}
 
@@ -475,8 +583,8 @@ def test_memmap_stay_open_bug(tmpdir):
     assert len(p.open_files()) <= len(orig_open)
 
 
-def test_masked_array_repr(tmpdir):
-    tmppath = os.path.join(str(tmpdir), "masked.asdf")
+def test_masked_array_repr(tmp_path):
+    tmppath = os.path.join(str(tmp_path), "masked.asdf")
 
     tree = {"array": np.arange(10), "masked": np.ma.array([1, 2, 3], mask=[False, True, False])}
 
@@ -486,8 +594,8 @@ def test_masked_array_repr(tmpdir):
         assert "masked array" in repr(ff.tree["masked"])
 
 
-def test_operations_on_ndarray_proxies(tmpdir):
-    tmppath = os.path.join(str(tmpdir), "test.asdf")
+def test_operations_on_ndarray_proxies(tmp_path):
+    tmppath = os.path.join(str(tmp_path), "test.asdf")
 
     tree = {"array": np.arange(10)}
 
@@ -508,13 +616,13 @@ def test_operations_on_ndarray_proxies(tmpdir):
         assert_array_equal(ff.tree["array"], x)
 
 
-def test_mask_datatype(tmpdir):
+def test_mask_datatype(tmp_path):
     content = """
-        arr: !core/ndarray-1.0.0
-            data: [1, 2, 3]
-            dtype: int32
-            mask: !core/ndarray-1.0.0
-                data: [true, true, false]
+arr: !core/ndarray-1.0.0
+    data: [1, 2, 3]
+    dtype: int32
+    mask: !core/ndarray-1.0.0
+        data: [true, true, false]
     """
     buff = helpers.yaml_to_asdf(content)
 
@@ -522,259 +630,283 @@ def test_mask_datatype(tmpdir):
         pass
 
 
-def test_invalid_mask_datatype(tmpdir):
+def test_invalid_mask_datatype(tmp_path):
     content = """
-        arr: !core/ndarray-1.0.0
-            data: [1, 2, 3]
-            dtype: int32
-            mask: !core/ndarray-1.0.0
-                data: ['a', 'b', 'c']
+arr: !core/ndarray-1.0.0
+    data: [1, 2, 3]
+    dtype: int32
+    mask: !core/ndarray-1.0.0
+        data: ['a', 'b', 'c']
     """
     buff = helpers.yaml_to_asdf(content)
 
-    with pytest.raises(jsonschema.ValidationError, match=r".* is not valid under any of the given schemas"), asdf.open(
-        buff,
+    with (
+        pytest.raises(ValidationError, match=r".* is not valid under any of the given schemas"),
+        asdf.open(
+            buff,
+        ),
     ):
         pass
 
 
-def test_ndim_validation(tmpdir):
+@with_custom_extension()
+def test_ndim_validation(tmp_path):
     content = """
-    obj: !<tag:nowhere.org:custom/ndim-1.0.0>
-        a: !core/ndarray-1.0.0
-           data: [1, 2, 3]
+obj: !<tag:nowhere.org:custom/ndim-1.0.0>
+    a: !core/ndarray-1.0.0
+       data: [1, 2, 3]
     """
     buff = helpers.yaml_to_asdf(content)
 
-    with pytest.raises(jsonschema.ValidationError, match=r"Wrong number of dimensions:.*"), asdf.open(
-        buff,
-        extensions=CustomExtension(),
+    with (
+        pytest.raises(ValidationError, match=r"Wrong number of dimensions:.*"),
+        asdf.open(
+            buff,
+        ),
     ):
         pass
 
     content = """
-    obj: !<tag:nowhere.org:custom/ndim-1.0.0>
-        a: !core/ndarray-1.0.0
-           data: [[1, 2, 3]]
+obj: !<tag:nowhere.org:custom/ndim-1.0.0>
+    a: !core/ndarray-1.0.0
+       data: [[1, 2, 3]]
     """
     buff = helpers.yaml_to_asdf(content)
 
-    with asdf.open(buff, extensions=CustomExtension()):
+    with asdf.open(buff):
         pass
 
     content = """
-    obj: !<tag:nowhere.org:custom/ndim-1.0.0>
-        a: !core/ndarray-1.0.0
-           shape: [1, 3]
-           data: [[1, 2, 3]]
+obj: !<tag:nowhere.org:custom/ndim-1.0.0>
+    a: !core/ndarray-1.0.0
+       shape: [1, 3]
+       data: [[1, 2, 3]]
     """
     buff = helpers.yaml_to_asdf(content)
 
-    with asdf.open(buff, extensions=CustomExtension()):
+    with asdf.open(buff):
         pass
 
     content = """
-    obj: !<tag:nowhere.org:custom/ndim-1.0.0>
-        b: !core/ndarray-1.0.0
-           data: [1, 2, 3]
+obj: !<tag:nowhere.org:custom/ndim-1.0.0>
+    b: !core/ndarray-1.0.0
+       data: [1, 2, 3]
     """
     buff = helpers.yaml_to_asdf(content)
 
-    with asdf.open(buff, extensions=CustomExtension()):
+    with asdf.open(buff):
         pass
 
     content = """
-    obj: !<tag:nowhere.org:custom/ndim-1.0.0>
-        b: !core/ndarray-1.0.0
-           data: [[1, 2, 3]]
+obj: !<tag:nowhere.org:custom/ndim-1.0.0>
+    b: !core/ndarray-1.0.0
+       data: [[1, 2, 3]]
     """
     buff = helpers.yaml_to_asdf(content)
 
-    with asdf.open(buff, extensions=CustomExtension()):
+    with asdf.open(buff):
         pass
 
     content = """
-    obj: !<tag:nowhere.org:custom/ndim-1.0.0>
-        b: !core/ndarray-1.0.0
-           data: [[[1, 2, 3]]]
+obj: !<tag:nowhere.org:custom/ndim-1.0.0>
+    b: !core/ndarray-1.0.0
+       data: [[[1, 2, 3]]]
     """
     buff = helpers.yaml_to_asdf(content)
 
-    with pytest.raises(jsonschema.ValidationError, match=r"Wrong number of dimensions:.*"), asdf.open(
-        buff,
-        extensions=CustomExtension(),
+    with (
+        pytest.raises(ValidationError, match=r"Wrong number of dimensions:.*"),
+        asdf.open(
+            buff,
+        ),
     ):
         pass
 
 
-def test_datatype_validation(tmpdir):
+@with_custom_extension()
+def test_datatype_validation(tmp_path):
     content = """
-    obj: !<tag:nowhere.org:custom/datatype-1.0.0>
-        a: !core/ndarray-1.0.0
-           data: [1, 2, 3]
-           datatype: float32
+obj: !<tag:nowhere.org:custom/datatype-1.0.0>
+    a: !core/ndarray-1.0.0
+       data: [1, 2, 3]
+       datatype: float32
     """
     buff = helpers.yaml_to_asdf(content)
 
-    with asdf.open(buff, extensions=CustomExtension()):
+    with asdf.open(buff):
         pass
 
     content = """
-    obj: !<tag:nowhere.org:custom/datatype-1.0.0>
-        a: !core/ndarray-1.0.0
-           data: [1, 2, 3]
+obj: !<tag:nowhere.org:custom/datatype-1.0.0>
+    a: !core/ndarray-1.0.0
+       data: [1, 2, 3]
+       datatype: float64
+    """
+    buff = helpers.yaml_to_asdf(content)
+
+    with (
+        pytest.raises(ValidationError, match=r"Can not safely cast from .* to .*"),
+        asdf.open(
+            buff,
+        ),
+    ):
+        pass
+
+    content = """
+obj: !<tag:nowhere.org:custom/datatype-1.0.0>
+    a: !core/ndarray-1.0.0
+       data: [1, 2, 3]
+       datatype: int16
+    """
+    buff = helpers.yaml_to_asdf(content)
+
+    with asdf.open(buff):
+        pass
+
+    content = """
+obj: !<tag:nowhere.org:custom/datatype-1.0.0>
+    b: !core/ndarray-1.0.0
+       data: [1, 2, 3]
+       datatype: int16
+    """
+    buff = helpers.yaml_to_asdf(content)
+
+    with (
+        pytest.raises(ValidationError, match=r"Expected datatype .*, got .*"),
+        asdf.open(
+            buff,
+        ),
+    ):
+        pass
+
+    content = """
+obj: !<tag:nowhere.org:custom/datatype-1.0.0>
+    a: !core/ndarray-1.0.0
+       data: [[1, 'a'], [2, 'b'], [3, 'c']]
+       datatype:
+         - name: a
+           datatype: int8
+         - name: b
+           datatype: ['ascii', 8]
+    """
+    buff = helpers.yaml_to_asdf(content)
+
+    with (
+        pytest.raises(ValidationError, match=r"Expected scalar datatype .*, got .*"),
+        asdf.open(
+            buff,
+        ),
+    ):
+        pass
+
+
+@with_custom_extension()
+def test_structured_datatype_validation(tmp_path):
+    content = """
+obj: !<tag:nowhere.org:custom/datatype-1.0.0>
+    c: !core/ndarray-1.0.0
+       data: [[1, 'a'], [2, 'b'], [3, 'c']]
+       datatype:
+         - name: a
+           datatype: int8
+         - name: b
+           datatype: ['ascii', 8]
+    """
+    buff = helpers.yaml_to_asdf(content)
+
+    with asdf.open(buff):
+        pass
+
+    content = """
+obj: !<tag:nowhere.org:custom/datatype-1.0.0>
+    c: !core/ndarray-1.0.0
+       data: [[1, 'a'], [2, 'b'], [3, 'c']]
+       datatype:
+         - name: a
+           datatype: int64
+         - name: b
+           datatype: ['ascii', 8]
+    """
+    buff = helpers.yaml_to_asdf(content)
+
+    with (
+        pytest.raises(ValidationError, match=r"Can not safely cast to expected datatype.*"),
+        asdf.open(
+            buff,
+        ),
+    ):
+        pass
+
+    content = """
+obj: !<tag:nowhere.org:custom/datatype-1.0.0>
+    c: !core/ndarray-1.0.0
+       data: [[1, 'a', 0], [2, 'b', 1], [3, 'c', 2]]
+       datatype:
+         - name: a
+           datatype: int8
+         - name: b
+           datatype: ['ascii', 8]
+         - name: c
            datatype: float64
     """
     buff = helpers.yaml_to_asdf(content)
 
-    with pytest.raises(jsonschema.ValidationError, match=r"Can not safely cast from .* to .*"), asdf.open(
-        buff,
-        extensions=CustomExtension(),
+    with (
+        pytest.raises(ValidationError, match=r"Mismatch in number of columns:.*"),
+        asdf.open(
+            buff,
+        ),
     ):
         pass
 
     content = """
-    obj: !<tag:nowhere.org:custom/datatype-1.0.0>
-        a: !core/ndarray-1.0.0
-           data: [1, 2, 3]
+obj: !<tag:nowhere.org:custom/datatype-1.0.0>
+    c: !core/ndarray-1.0.0
+       data: [1, 2, 3]
+    """
+    buff = helpers.yaml_to_asdf(content)
+
+    with (
+        pytest.raises(ValidationError, match=r"Expected structured datatype.*"),
+        asdf.open(
+            buff,
+        ),
+    ):
+        pass
+
+    content = """
+obj: !<tag:nowhere.org:custom/datatype-1.0.0>
+    d: !core/ndarray-1.0.0
+       data: [[1, 'a'], [2, 'b'], [3, 'c']]
+       datatype:
+         - name: a
+           datatype: int8
+         - name: b
+           datatype: ['ascii', 8]
+    """
+    buff = helpers.yaml_to_asdf(content)
+
+    with (
+        pytest.raises(ValidationError, match=r"Expected datatype .*, got .*"),
+        asdf.open(
+            buff,
+        ),
+    ):
+        pass
+
+    content = """
+obj: !<tag:nowhere.org:custom/datatype-1.0.0>
+    d: !core/ndarray-1.0.0
+       data: [[1, 'a'], [2, 'b'], [3, 'c']]
+       datatype:
+         - name: a
            datatype: int16
+         - name: b
+           datatype: ['ascii', 16]
     """
     buff = helpers.yaml_to_asdf(content)
 
-    with asdf.open(buff, extensions=CustomExtension()):
-        pass
-
-    content = """
-    obj: !<tag:nowhere.org:custom/datatype-1.0.0>
-        b: !core/ndarray-1.0.0
-           data: [1, 2, 3]
-           datatype: int16
-    """
-    buff = helpers.yaml_to_asdf(content)
-
-    with pytest.raises(jsonschema.ValidationError, match=r"Expected datatype .*, got .*"), asdf.open(
-        buff,
-        extensions=CustomExtension(),
-    ):
-        pass
-
-    content = """
-    obj: !<tag:nowhere.org:custom/datatype-1.0.0>
-        a: !core/ndarray-1.0.0
-           data: [[1, 'a'], [2, 'b'], [3, 'c']]
-           datatype:
-             - name: a
-               datatype: int8
-             - name: b
-               datatype: ['ascii', 8]
-    """
-    buff = helpers.yaml_to_asdf(content)
-
-    with pytest.raises(jsonschema.ValidationError, match=r"Expected scalar datatype .*, got .*"), asdf.open(
-        buff,
-        extensions=CustomExtension(),
-    ):
-        pass
-
-
-def test_structured_datatype_validation(tmpdir):
-    content = """
-    obj: !<tag:nowhere.org:custom/datatype-1.0.0>
-        c: !core/ndarray-1.0.0
-           data: [[1, 'a'], [2, 'b'], [3, 'c']]
-           datatype:
-             - name: a
-               datatype: int8
-             - name: b
-               datatype: ['ascii', 8]
-    """
-    buff = helpers.yaml_to_asdf(content)
-
-    with asdf.open(buff, extensions=CustomExtension()):
-        pass
-
-    content = """
-    obj: !<tag:nowhere.org:custom/datatype-1.0.0>
-        c: !core/ndarray-1.0.0
-           data: [[1, 'a'], [2, 'b'], [3, 'c']]
-           datatype:
-             - name: a
-               datatype: int64
-             - name: b
-               datatype: ['ascii', 8]
-    """
-    buff = helpers.yaml_to_asdf(content)
-
-    with pytest.raises(jsonschema.ValidationError, match=r"Can not safely cast to expected datatype.*"), asdf.open(
-        buff,
-        extensions=CustomExtension(),
-    ):
-        pass
-
-    content = """
-    obj: !<tag:nowhere.org:custom/datatype-1.0.0>
-        c: !core/ndarray-1.0.0
-           data: [[1, 'a', 0], [2, 'b', 1], [3, 'c', 2]]
-           datatype:
-             - name: a
-               datatype: int8
-             - name: b
-               datatype: ['ascii', 8]
-             - name: c
-               datatype: float64
-    """
-    buff = helpers.yaml_to_asdf(content)
-
-    with pytest.raises(jsonschema.ValidationError, match=r"Mismatch in number of columns:.*"), asdf.open(
-        buff,
-        extensions=CustomExtension(),
-    ):
-        pass
-
-    content = """
-    obj: !<tag:nowhere.org:custom/datatype-1.0.0>
-        c: !core/ndarray-1.0.0
-           data: [1, 2, 3]
-    """
-    buff = helpers.yaml_to_asdf(content)
-
-    with pytest.raises(jsonschema.ValidationError, match=r"Expected structured datatype.*"), asdf.open(
-        buff,
-        extensions=CustomExtension(),
-    ):
-        pass
-
-    content = """
-    obj: !<tag:nowhere.org:custom/datatype-1.0.0>
-        d: !core/ndarray-1.0.0
-           data: [[1, 'a'], [2, 'b'], [3, 'c']]
-           datatype:
-             - name: a
-               datatype: int8
-             - name: b
-               datatype: ['ascii', 8]
-    """
-    buff = helpers.yaml_to_asdf(content)
-
-    with pytest.raises(jsonschema.ValidationError, match=r"Expected datatype .*, got .*"), asdf.open(
-        buff,
-        extensions=CustomExtension(),
-    ):
-        pass
-
-    content = """
-    obj: !<tag:nowhere.org:custom/datatype-1.0.0>
-        d: !core/ndarray-1.0.0
-           data: [[1, 'a'], [2, 'b'], [3, 'c']]
-           datatype:
-             - name: a
-               datatype: int16
-             - name: b
-               datatype: ['ascii', 16]
-    """
-    buff = helpers.yaml_to_asdf(content)
-
-    with asdf.open(buff, extensions=CustomExtension()):
+    with asdf.open(buff):
         pass
 
 
@@ -788,9 +920,9 @@ def test_string_inline():
 
 def test_inline_shape_mismatch():
     content = """
-    arr: !core/ndarray-1.0.0
-      data: [1, 2, 3]
-      shape: [2]
+arr: !core/ndarray-1.0.0
+  data: [1, 2, 3]
+  shape: [2]
     """
 
     buff = helpers.yaml_to_asdf(content)
@@ -798,68 +930,45 @@ def test_inline_shape_mismatch():
         pass
 
 
-@pytest.mark.xfail(reason="NDArrays with dtype=object are not currently supported")
-def test_simple_object_array(tmpdir):
-    # See https://github.com/asdf-format/asdf/issues/383 for feature
-    # request
-    dictdata = np.empty((3, 3), dtype=object)
-    for i, _ in enumerate(dictdata.flat):
-        dictdata.flat[i] = {"foo": i * 42, "bar": i**2}
-
-    helpers.assert_roundtrip_tree({"bizbaz": dictdata}, tmpdir)
-
-
-@pytest.mark.xfail(reason="NDArrays with dtype=object are not currently supported")
-def test_tagged_object_array(tmpdir):
-    # See https://github.com/asdf-format/asdf/issues/383 for feature
-    # request
-    quantity = pytest.importorskip("astropy.units.quantity")
-
-    objdata = np.empty((3, 3), dtype=object)
-    for i, _ in enumerate(objdata.flat):
-        objdata.flat[i] = quantity.Quantity(i, "angstrom")
-
-    helpers.assert_roundtrip_tree({"bizbaz": objdata}, tmpdir)
-
-
-def test_broadcasted_array(tmpdir):
+def test_broadcasted_array(tmp_path):
     attrs = np.broadcast_arrays(np.array([10, 20]), np.array(10), np.array(10))
     tree = {"one": attrs[1]}  # , 'two': attrs[1], 'three': attrs[2]}
-    helpers.assert_roundtrip_tree(tree, tmpdir)
+    with roundtrip(tree) as af:
+        assert_array_equal(tree["one"], af["one"])
 
 
-def test_broadcasted_offset_array(tmpdir):
+def test_broadcasted_offset_array(tmp_path):
     base = np.arange(10)
     offset = base[5:]
     broadcasted = np.broadcast_to(offset, (4, 5))
     tree = {"broadcasted": broadcasted}
-    helpers.assert_roundtrip_tree(tree, tmpdir)
+    with roundtrip(tree) as af:
+        assert_array_equal(tree["broadcasted"], af["broadcasted"])
 
 
-def test_non_contiguous_base_array(tmpdir):
+def test_non_contiguous_base_array(tmp_path):
     base = np.arange(60).reshape(5, 4, 3).transpose(2, 0, 1) * 1
     contiguous = base.transpose(1, 2, 0)
     tree = {"contiguous": contiguous}
-    helpers.assert_roundtrip_tree(tree, tmpdir)
+    with roundtrip(tree) as af:
+        assert_array_equal(tree["contiguous"], af["contiguous"])
 
 
-def test_fortran_order(tmpdir):
+def test_fortran_order(tmp_path):
     array = np.array([[11, 12, 13], [21, 22, 23]], order="F", dtype=np.int64)
     tree = {"data": array}
 
-    def check_f_order(t):
-        assert t["data"].flags.fortran
-        assert np.all(np.isclose(array, t["data"]))
+    with roundtrip(tree) as af:
+        assert af["data"].flags.fortran
+        assert np.all(np.isclose(array, af["data"]))
 
-    def check_raw_yaml(content):
+    with roundtrip(tree, raw=True) as content:
         tree = yaml.safe_load(re.sub(rb"!core/\S+", b"", content))
         assert tree["data"]["strides"] == [8, 16]
 
-    helpers.assert_roundtrip_tree(tree, tmpdir, asdf_check_func=check_f_order, raw_yaml_check_func=check_raw_yaml)
 
-
-def test_memmap_write(tmpdir):
-    tmpfile = str(tmpdir.join("data.asdf"))
+def test_memmap_write(tmp_path):
+    tmpfile = str(tmp_path / "data.asdf")
     tree = {"data": np.zeros(100)}
 
     with asdf.AsdfFile(tree) as af:
@@ -879,8 +988,8 @@ def test_memmap_write(tmpdir):
         assert af["data"][0] == 42
 
 
-def test_readonly(tmpdir):
-    tmpfile = str(tmpdir.join("data.asdf"))
+def test_readonly(tmp_path):
+    tmpfile = str(tmp_path / "data.asdf")
     tree = {"data": np.ndarray(100)}
 
     with asdf.AsdfFile(tree) as af:
@@ -910,8 +1019,8 @@ def test_readonly(tmpdir):
         af["data"][0] = 42
 
 
-def test_readonly_inline(tmpdir):
-    tmpfile = str(tmpdir.join("data.asdf"))
+def test_readonly_inline(tmp_path):
+    tmpfile = str(tmp_path / "data.asdf")
     tree = {"data": np.ndarray(100)}
 
     with asdf.AsdfFile(tree) as af:
@@ -926,8 +1035,8 @@ def test_readonly_inline(tmpdir):
 # Confirm that NDArrayType's internal array is regenerated
 # following an update.
 @pytest.mark.parametrize("pad_blocks", [True, False])
-def test_block_data_change(pad_blocks, tmpdir):
-    tmpfile = str(tmpdir.join("data.asdf"))
+def test_block_data_change(pad_blocks, tmp_path):
+    tmpfile = str(tmp_path / "data.asdf")
     tree = {"data": np.zeros(10, dtype="uint8")}
     with asdf.AsdfFile(tree) as af:
         af.write_to(tmpfile, pad_blocks=pad_blocks)
@@ -964,3 +1073,15 @@ def test_problematic_class_attributes(tmp_path):
 
         with pytest.raises(AttributeError, match=r".* object has no attribute 'version'"):
             af["arr"].version
+
+
+def test_shape_does_not_load_array(tmp_path):
+    file_path = tmp_path / "test.asdf"
+    with asdf.AsdfFile() as af:
+        af["arr"] = np.arange(100)
+        af.write_to(file_path)
+
+    with asdf.open(file_path, lazy_load=True) as af:
+        assert af["arr"]._array is None
+        assert af["arr"].shape == (100,)
+        assert af["arr"]._array is None
