@@ -1,10 +1,11 @@
+import fractions
+
 import pytest
 from packaging.specifiers import SpecifierSet
+from yaml.representer import RepresenterError
 
 from asdf import AsdfFile, config_context
-from asdf._tests._helpers import assert_extension_correctness
-from asdf._types import CustomType
-from asdf.exceptions import AsdfDeprecationWarning, ValidationError
+from asdf.exceptions import AsdfWarning, ValidationError
 from asdf.extension import (
     Compressor,
     Converter,
@@ -15,30 +16,9 @@ from asdf.extension import (
     ManifestExtension,
     TagDefinition,
     Validator,
-    get_cached_asdf_extension_list,
     get_cached_extension_manager,
 )
-from asdf.extension._legacy import AsdfExtension, BuiltinExtension
-
-
-def test_builtin_extension():
-    extension = BuiltinExtension()
-    with pytest.warns(AsdfDeprecationWarning, match="assert_extension_correctness is deprecated.*"):
-        assert_extension_correctness(extension)
-
-
-with pytest.warns(AsdfDeprecationWarning, match=".*subclasses the deprecated CustomType.*"):
-
-    class LegacyType(dict, CustomType):
-        organization = "somewhere.org"
-        name = "test"
-        version = "1.0.0"
-
-
-class LegacyExtension:
-    types = [LegacyType]
-    tag_mapping = [("tag:somewhere.org/", "http://somewhere.org/{tag_suffix}")]
-    url_mapping = [("http://somewhere.org/", "http://somewhere.org/{url_suffix}.yaml")]
+from asdf.testing.helpers import roundtrip_object
 
 
 class MinimumExtension:
@@ -149,6 +129,10 @@ class FooType:
     pass
 
 
+class SubFooType(FooType):
+    pass
+
+
 class BarType:
     pass
 
@@ -163,7 +147,7 @@ def test_extension_proxy_maybe_wrap():
     assert proxy.delegate is extension
     assert ExtensionProxy.maybe_wrap(proxy) is proxy
 
-    with pytest.raises(TypeError, match=r"Extension must implement the Extension or AsdfExtension interface"):
+    with pytest.raises(TypeError, match=r"Extension must implement the Extension interface"):
         ExtensionProxy.maybe_wrap(object())
 
 
@@ -173,7 +157,6 @@ def test_extension_proxy():
     proxy = ExtensionProxy(extension)
 
     assert isinstance(proxy, Extension)
-    assert isinstance(proxy, AsdfExtension)
 
     assert proxy.extension_uri == "asdf://somewhere.org/extensions/minimum-1.0"
     assert proxy.legacy_class_names == set()
@@ -242,7 +225,7 @@ def test_extension_proxy():
     assert proxy.class_name == "asdf._tests.test_extension.FullExtension"
 
     # Should fail when the input is not one of the two extension interfaces:
-    with pytest.raises(TypeError, match=r"Extension must implement the Extension or AsdfExtension interface"):
+    with pytest.raises(TypeError, match=r"Extension must implement the Extension interface"):
         ExtensionProxy(object)
 
     # Should fail with a bad converter:
@@ -316,25 +299,6 @@ def test_extension_proxy_tags():
     assert proxy.converters[0].tags == [foo_tag_uri]
 
 
-def test_extension_proxy_legacy():
-    extension = LegacyExtension()
-    proxy = ExtensionProxy(extension, package_name="foo", package_version="1.2.3")
-
-    assert proxy.extension_uri is None
-    assert proxy.legacy_class_names == {"asdf._tests.test_extension.LegacyExtension"}
-    assert proxy.asdf_standard_requirement == SpecifierSet()
-    assert proxy.converters == []
-    assert proxy.tags == []
-    assert proxy.types == [LegacyType]
-    assert proxy.tag_mapping == LegacyExtension.tag_mapping
-    assert proxy.url_mapping == LegacyExtension.url_mapping
-    assert proxy.delegate is extension
-    assert proxy.legacy is True
-    assert proxy.package_name == "foo"
-    assert proxy.package_version == "1.2.3"
-    assert proxy.class_name == "asdf._tests.test_extension.LegacyExtension"
-
-
 def test_extension_proxy_hash_and_eq():
     extension = MinimumExtension()
     proxy1 = ExtensionProxy(extension)
@@ -357,11 +321,6 @@ def test_extension_proxy_repr():
     assert "package: (none)" in repr(proxy)
     assert "legacy: False" in repr(proxy)
 
-    proxy = ExtensionProxy(LegacyExtension(), package_name="foo", package_version="1.2.3")
-    assert "class: asdf._tests.test_extension.LegacyExtension" in repr(proxy)
-    assert "package: foo==1.2.3" in repr(proxy)
-    assert "legacy: True" in repr(proxy)
-
 
 def test_extension_manager():
     converter1 = FullConverter(
@@ -412,6 +371,7 @@ def test_extension_manager():
     assert manager.handles_tag("asdf://somewhere.org/extensions/full/tags/baz-1.0") is True
 
     assert manager.handles_type(FooType) is True
+    assert manager.handles_type(SubFooType) is False
     # This should return True even though BarType was listed
     # as string class name:
     assert manager.handles_type(BarType) is True
@@ -438,6 +398,8 @@ def test_extension_manager():
     assert manager.get_converter_for_type(BazType).delegate is converter2
     with pytest.raises(KeyError, match=r"\"No support available for Python type .*\""):
         manager.get_converter_for_type(object)
+    with pytest.raises(KeyError, match=r"\"No support available for Python type .*\""):
+        manager.get_converter_for_type(SubFooType)
 
 
 def test_get_cached_extension_manager():
@@ -462,9 +424,6 @@ def test_tag_definition():
 
     assert "URI: asdf://somewhere.org/extensions/foo/tags/foo-1.0" in repr(tag_def)
 
-    with pytest.warns(AsdfDeprecationWarning, match=r"The .* property is deprecated.*"):
-        assert tag_def.schema_uri == "asdf://somewhere.org/extensions/foo/schemas/foo-1.0"
-
     tag_def = TagDefinition(
         "asdf://somewhere.org/extensions/foo/tags/foo-1.0",
         schema_uris=[
@@ -479,11 +438,6 @@ def test_tag_definition():
         "asdf://somewhere.org/extensions/foo/schemas/foo-1.0",
         "asdf://somewhere.org/extensions/foo/schemas/base-1.0",
     ]
-    with pytest.warns(AsdfDeprecationWarning, match=r"The .* property is deprecated.*"), pytest.raises(
-        RuntimeError,
-        match=r"Cannot use .* when multiple schema URIs are present",
-    ):
-        tag_def.schema_uri
 
     with pytest.raises(ValueError, match=r"URI patterns are not permitted in TagDefinition"):
         TagDefinition("asdf://somewhere.org/extensions/foo/tags/foo-*")
@@ -502,19 +456,6 @@ def test_converter():
 
     assert issubclass(ConverterNoSubclass, Converter)
 
-    class ConverterWithSubclass(Converter):
-        tags = []
-        types = []
-
-        def to_yaml_tree(self, *args):
-            pass
-
-        def from_yaml_tree(self, *args):
-            pass
-
-    # Confirm the behavior of the default select_tag implementation
-    assert ConverterWithSubclass().select_tag(object(), ["tag1", "tag2"], object()) == "tag1"
-
 
 def test_converter_proxy():
     # Test the minimum set of converter methods:
@@ -601,17 +542,41 @@ def test_converter_proxy():
 
     # Should fail because types must instances of type:
     with pytest.raises(TypeError, match=r"Converter property .* must contain str or type values"):
-        ConverterProxy(MinimumConverter(types=[object()]), extension)
+        # as the code will ignore types if no relevant tags are found
+        # include a tag from this extension to make sure the proxy considers
+        # the types
+        ConverterProxy(MinimumConverter(tags=[extension.tags[0].tag_uri], types=[object()]), extension)
 
 
-def test_get_cached_asdf_extension_list():
-    extension = LegacyExtension()
-    with pytest.warns(AsdfDeprecationWarning, match="get_cached_asdf_extension_list is deprecated"):
-        extension_list = get_cached_asdf_extension_list([extension])
-    with pytest.warns(AsdfDeprecationWarning, match="get_cached_asdf_extension_list is deprecated"):
-        assert get_cached_asdf_extension_list([extension]) is extension_list
-    with pytest.warns(AsdfDeprecationWarning, match="get_cached_asdf_extension_list is deprecated"):
-        assert get_cached_asdf_extension_list([LegacyExtension()]) is not extension_list
+def test_converter_subclass_with_no_supported_tags():
+    """
+    Adding a Converter to an Extension that doesn't list support for the tags
+    associated with the Converter should result in a failure to convert.
+    """
+
+    class Foo:
+        pass
+
+    class FooConverterWithSubclass(Converter):
+        tags = ["asdf://somewhere.org/tags/foo-1.0.0"]
+        types = [Foo]
+
+        def to_yaml_tree(self, *args):
+            pass
+
+        def from_yaml_tree(self, *args):
+            pass
+
+    class FooExtension(Extension):
+        tags = []
+        converters = [FooConverterWithSubclass()]
+        extension_uri = "asdf://somewhere.org/extensions/foo-1.0.0"
+
+    tree = {"obj": Foo()}
+    with config_context() as cfg:
+        cfg.add_extension(FooExtension())
+        with pytest.raises(RepresenterError, match=r"cannot represent an object"):
+            roundtrip_object(tree)
 
 
 def test_manifest_extension():
@@ -764,3 +729,215 @@ def test_validator():
             af["foo"] = "bar"
             with pytest.raises(ValidationError, match=r"Node was doomed to fail"):
                 af.validate()
+
+
+def test_converter_deferral():
+    class Bar:
+        def __init__(self, value):
+            self.value = value
+
+    class Foo(Bar):
+        pass
+
+    class Baz(Bar):
+        pass
+
+    class FooConverter:
+        tags = []
+        types = [Foo]
+
+        def select_tag(self, *args):
+            return None
+
+        def to_yaml_tree(self, obj, tag, ctx):
+            # convert Foo instance to Bar
+            return Bar(obj.value)
+
+        def from_yaml_tree(self, node, tag, ctx):
+            raise NotImplementedError()
+
+    class BarConverter:
+        tags = ["asdf://somewhere.org/tags/bar"]
+        types = [Bar]
+
+        def to_yaml_tree(self, obj, tag, ctx):
+            return {"value": obj.value}
+
+        def from_yaml_tree(self, node, tag, ctx):
+            return Bar(node["value"])
+
+    class BazConverter:
+        tags = []
+        types = [Baz]
+
+        def select_tag(self, *args):
+            return None
+
+        def to_yaml_tree(self, obj, tag, ctx):
+            return Foo(obj.value)
+
+        def from_yaml_tree(self, node, tag, ctx):
+            raise NotImplementedError()
+
+    extension = FullExtension(converters=[FooConverter(), BarConverter(), BazConverter()], tags=BarConverter.tags)
+    with config_context() as config:
+        config.add_extension(extension)
+
+        foo = Foo(26)
+        bar = Bar(42)
+        baz = Baz(720)
+
+        bar_rt = roundtrip_object(bar)
+        assert isinstance(bar_rt, Bar)
+        assert bar_rt.value == bar.value
+
+        foo_rt = roundtrip_object(foo)
+        assert isinstance(foo_rt, Bar)
+        assert foo_rt.value == foo.value
+
+        baz_rt = roundtrip_object(baz)
+        assert isinstance(baz_rt, Bar)
+        assert baz_rt.value == baz.value
+
+
+def test_converter_loop():
+    class Bar:
+        def __init__(self, value):
+            self.value = value
+
+    class Foo(Bar):
+        pass
+
+    class Baz(Bar):
+        pass
+
+    class FooConverter:
+        tags = []
+        types = [Foo]
+
+        def select_tag(self, *args):
+            return None
+
+        def to_yaml_tree(self, obj, tag, ctx):
+            return Bar(obj.value)
+
+        def from_yaml_tree(self, node, tag, ctx):
+            raise NotImplementedError()
+
+    class BarConverter:
+        tags = []
+        types = [Bar]
+
+        def select_tag(self, *args):
+            return None
+
+        def to_yaml_tree(self, obj, tag, ctx):
+            return Baz(obj.value)
+
+        def from_yaml_tree(self, node, tag, ctx):
+            raise NotImplementedError()
+
+    class BazConverter:
+        tags = []
+        types = [Baz]
+
+        def select_tag(self, *args):
+            return None
+
+        def to_yaml_tree(self, obj, tag, ctx):
+            return Foo(obj.value)
+
+        def from_yaml_tree(self, node, tag, ctx):
+            raise NotImplementedError()
+
+    extension = FullExtension(converters=[FooConverter(), BarConverter(), BazConverter()])
+    with config_context() as config:
+        config.add_extension(extension)
+
+        for typ in (Foo, Bar, Baz):
+            obj = typ(42)
+            with pytest.raises(TypeError, match=r"Conversion cycle detected"):
+                roundtrip_object(obj)
+
+
+@pytest.mark.parametrize("is_subclass", [True, False])
+@pytest.mark.parametrize("indirect", [True, False])
+def test_warning_or_error_for_default_select_tag(is_subclass, indirect):
+    class Foo:
+        pass
+
+    ParentClass = Converter if is_subclass else object
+
+    if indirect:
+
+        class IntermediateClass(ParentClass):
+            pass
+
+        ParentClass = IntermediateClass
+
+    class FooConverter(ParentClass):
+        tags = ["asdf://somewhere.org/tags/foo-*"]
+        types = [Foo]
+
+        def to_yaml_tree(self, obj, tag, ctx):
+            return {}
+
+        def from_yaml_tree(self, node, tag, ctx):
+            return Foo()
+
+    tags = [
+        "asdf://somewhere.org/tags/foo-1.0.0",
+        "asdf://somewhere.org/tags/foo-2.0.0",
+    ]
+    extension = FullExtension(converters=[FooConverter()], tags=tags)
+    ctx_type = pytest.warns if is_subclass else pytest.raises
+    exception_class = AsdfWarning if is_subclass else RuntimeError
+    with config_context() as config:
+        with ctx_type(exception_class, match="Converter handles multiple tags"):
+            config.add_extension(extension)
+
+
+def test_reference_cycle():
+    class FractionWithInverse(fractions.Fraction):
+        def __init__(self, *args, **kwargs):
+            self._inverse = None
+
+        @property
+        def inverse(self):
+            return self._inverse
+
+        @inverse.setter
+        def inverse(self, value):
+            self._inverse = value
+
+    class FractionWithInverseConverter:
+        tags = ["asdf://example.com/fractions/tags/fraction-1.0.0"]
+        types = [FractionWithInverse]
+
+        def to_yaml_tree(self, obj, tag, ctx):
+            return {
+                "numerator": obj.numerator,
+                "denominator": obj.denominator,
+                "inverse": obj.inverse,
+            }
+
+        def from_yaml_tree(self, node, tag, ctx):
+            obj = FractionWithInverse(node["numerator"], node["denominator"])
+            yield obj
+            obj.inverse = node["inverse"]
+
+    class FractionWithInverseExtension:
+        tags = FractionWithInverseConverter.tags
+        converters = [FractionWithInverseConverter()]
+        extension_uri = "asdf://example.com/fractions/extensions/fraction-1.0.0"
+
+    with config_context() as cfg:
+        cfg.add_extension(FractionWithInverseExtension())
+
+        f1 = FractionWithInverse(3, 5)
+        f2 = FractionWithInverse(5, 3)
+        f1.inverse = f2
+        f2.inverse = f1
+
+        read_f1 = roundtrip_object(f1)
+        assert read_f1.inverse.inverse is read_f1
