@@ -78,12 +78,12 @@ class ObjectWithInfoSupport:
         self.patt = {}
         for key in kw:
             if re.search("^S_", key):
-                if type(kw[key]) != str:
+                if not isinstance(kw[key], str):
                     msg = "S_ pattern object must be a string"
                     raise ValueError(msg)
                 self.patt[key] = kw[key]
             if re.search("^I_", key):
-                if type(kw[key]) != int:
+                if not isinstance(kw[key], int):
                     msg = "I_ pattern object must be an int"
                     raise ValueError(msg)
                 self.patt[key] = kw[key]
@@ -633,6 +633,9 @@ class RecursiveObjectWithInfoSupport:
     def __asdf_traverse__(self):
         return {"the_meaning": self.the_meaning, "clown": self.clown, "recursive": self.recursive}
 
+    def __str__(self):
+        return "rec ref"
+
 
 def test_recursive_info_object_support(capsys, tmp_path):
     tempdir = pathlib.Path(tempfile.mkdtemp())
@@ -662,3 +665,33 @@ def test_search():
 
     result = af.search(value="hello")
     assert result.node == "hello"
+
+
+def test_info_str(capsys):
+    class BadStr:
+        def __str__(self):
+            raise Exception()
+
+    class NewlineStr:
+        def __str__(self):
+            return "a\nb"
+
+    class CarriageReturnStr:
+        def __str__(self):
+            return "a\rb"
+
+    class NiceStr:
+        def __str__(self):
+            return "nice"
+
+    af = asdf.AsdfFile()
+    af["a"] = BadStr()
+    af["b"] = NewlineStr()
+    af["c"] = CarriageReturnStr()
+    af["d"] = NiceStr()
+    af.info()
+    captured = capsys.readouterr()
+    assert "(BadStr)\n" in captured.out
+    assert "(NewlineStr)\n" in captured.out
+    assert "(CarriageReturnStr)\n" in captured.out
+    assert "(NiceStr): nice\n" in captured.out
