@@ -6,6 +6,7 @@ import argparse
 import sys
 
 import jmespath
+import numpy as np
 from numpy import array_equal
 
 try:
@@ -30,9 +31,8 @@ except ImportError:
         RESET = ""
 
 import asdf
+from asdf.extension._serialization_context import BlockAccess
 from asdf.tagged import Tagged
-from asdf.tags.core.ndarray import NDArrayType
-from asdf.util import human_list
 
 from .main import Command
 
@@ -103,8 +103,9 @@ class ArrayNode:
     of but not necessarily display. This allows the diff output to be
     cleaner."""
 
-    def __init__(self, name):
+    def __init__(self, name, index):
         self.name = name
+        self.index = index
 
     def __hash__(self):
         return hash(self.name)
@@ -190,7 +191,7 @@ def print_in_tree(diff_ctx, node_list, thing, other, use_marker=False, last_was_
     # If tree element is list, recursively print list contents
     if isinstance(thing, list):
         for i, subthing in enumerate(thing):
-            key = ArrayNode(f"{node_list[-1]}_{i}")
+            key = ArrayNode(f"{node_list[-1]}_{i}", i)
             last_was_list = print_in_tree(
                 diff_ctx,
                 [*node_list, key],
@@ -244,6 +245,44 @@ def print_dict_diff(diff_ctx, tree, node_list, keys, other):
         print_in_tree(diff_ctx, nodes, key_, other, use_marker=use_marker)
 
 
+def _load_array(asdf_file, array_dict):
+    # the array_dict may not be tagged if the array is inline
+    # in this case just use what's in "data"
+    if not hasattr(array_dict, "_tag"):
+        return array_dict["data"]
+    conv = asdf_file.extension_manager.get_converter_for_type(np.ndarray)
+    sctx = asdf_file._create_serialization_context(BlockAccess.READ)
+    return conv.from_yaml_tree(array_dict, array_dict._tag, sctx)
+
+
+def _human_list(line, separator="and"):
+    """
+    Formats a list for human readability.
+
+    Parameters
+    ----------
+    line : sequence
+        A sequence of strings
+
+    separator : string, optional
+        The word to use between the last two entries.  Default:
+        ``"and"``.
+
+    Returns
+    -------
+    formatted_list : string
+
+    Examples
+    --------
+    >>> _human_list(["vanilla", "strawberry", "chocolate"], "or")
+    'vanilla, strawberry or chocolate'
+    """
+    if len(line) == 1:
+        return line[0]
+
+    return ", ".join(line[:-1]) + " " + separator + " " + line[-1]
+
+
 def compare_ndarrays(diff_ctx, array0, array1, keys):
     """Compares two ndarray objects"""
     if isinstance(array0, list):
@@ -259,15 +298,16 @@ def compare_ndarrays(diff_ctx, array0, array1, keys):
         if array0.get(field) != array1.get(field):
             differences.append(field)
 
-    array0 = NDArrayType.from_tree(array0, diff_ctx.asdf0)
-    array1 = NDArrayType.from_tree(array1, diff_ctx.asdf1)
-    if not array_equal(array0, array1):
+    value0 = _load_array(diff_ctx.asdf0, array0)
+    value1 = _load_array(diff_ctx.asdf1, array1)
+
+    if not array_equal(value0, value1):
         differences.append("contents")
 
     if differences:
-        prefix = "  " * (len(keys) + 1)
-        msg = f"ndarrays differ by {human_list(differences)}"
-        diff_ctx.iostream.write(prefix + RED + msg + RESET_NEWLINE)
+        msg = f"ndarrays differ by {_human_list(differences)}"
+        print_in_tree(diff_ctx, keys, msg, False, ignore_lwl=True)
+        print_in_tree(diff_ctx, keys, msg, True, ignore_lwl=True)
 
 
 def both_are_ndarrays(tree0, tree1):
@@ -309,7 +349,7 @@ def compare_trees(diff_ctx, tree0, tree1, keys=None):
         compare_dicts(diff_ctx, tree0, tree1, keys)
     elif isinstance(tree0, list) and isinstance(tree1, list):
         for i, (obj0, obj1) in enumerate(zip(tree0, tree1)):
-            key = ArrayNode(f"item_{i}")
+            key = ArrayNode(f"item_{i}", i)
             compare_trees(diff_ctx, obj0, obj1, [*keys, key])
     else:
         compare_objects(diff_ctx, tree0, tree1, keys)
@@ -338,10 +378,13 @@ def diff(filenames, minimal, iostream=sys.stdout, ignore=None):
     ignore_expressions = [] if ignore is None else [jmespath.compile(e) for e in ignore]
 
     try:
-        with asdf.open(filenames[0], _force_raw_types=True) as asdf0, asdf.open(
-            filenames[1],
-            _force_raw_types=True,
-        ) as asdf1:
+        with (
+            asdf.open(filenames[0], _force_raw_types=True) as asdf0,
+            asdf.open(
+                filenames[1],
+                _force_raw_types=True,
+            ) as asdf1,
+        ):
             ignore_ids = set()
             for expression in ignore_expressions:
                 for tree in [asdf0.tree, asdf1.tree]:
