@@ -16,8 +16,9 @@ import tempfile
 import yaml
 
 from asdf import constants, generic_io, schema, util
-from asdf.asdf import AsdfFile, open_asdf
-from asdf.block import BlockManager
+from asdf._asdf import AsdfFile
+from asdf._block import io as bio
+from asdf._block.exceptions import BlockIndexError
 
 from .main import Command
 
@@ -130,23 +131,41 @@ def write_edited_yaml_larger(path, new_content, version):
             pad_length = util.calculate_padding(len(new_content), True, fd.block_size)
             fd.fast_forward(pad_length)
 
-            with generic_io.get_file(path) as original_fd:
-                # Consume the file up to the first block, which must exist
-                # as a precondition to using this method.
-                original_fd.seek_until(
-                    constants.BLOCK_MAGIC,
-                    len(constants.BLOCK_MAGIC),
-                )
-                ctx = AsdfFile(version=version)
-                blocks = BlockManager(ctx, copy_arrays=False, lazy_load=False)
-                blocks.read_internal_blocks(original_fd, past_magic=True, validate_checksums=False)
-                blocks.finish_reading_internal_blocks()
-                blocks.write_internal_blocks_serial(fd)
-                blocks.write_block_index(fd, ctx)
-                blocks.close()
+            # now copy over ASDF block contents
 
-            # the file needs to be closed here to release all memmaps
-            original_fd.close()
+            with generic_io.get_file(path) as original_fd:
+                original_fd.seek_until(constants.BLOCK_MAGIC, len(constants.BLOCK_MAGIC))
+                old_first_block_offset = original_fd.tell() - len(constants.BLOCK_MAGIC)
+                new_first_block_offset = fd.tell()
+
+                # check if the original file has a block index which we will need to update
+                # as we're moving the blocks
+                block_index_offset = bio.find_block_index(original_fd)
+                if block_index_offset is None:
+                    block_index = None
+                    original_fd.seek(0, generic_io.SEEK_END)
+                    blocks_end = original_fd.tell()
+                else:
+                    blocks_end = block_index_offset
+                    try:
+                        block_index = bio.read_block_index(original_fd, block_index_offset)
+                    except BlockIndexError:
+                        # the original index was invalid
+                        block_index = None
+
+                # copy over blocks byte-for-byte from old_first_block_offset to block_index_offset
+                original_fd.seek(old_first_block_offset)
+                block_size = min(fd.block_size, original_fd.block_size)
+                n_bytes = blocks_end - old_first_block_offset
+                for offset in range(0, n_bytes, block_size):
+                    this_size = min(block_size, n_bytes - offset)
+                    fd.write(original_fd.read(this_size))
+
+                # update index
+                if block_index is not None:
+                    offset = new_first_block_offset - old_first_block_offset
+                    updated_block_index = [i + offset for i in block_index]
+                    bio.write_block_index(fd, updated_block_index)
 
         # Swap in the new version of the file atomically:
         shutil.copy(temp_file.name, path)
@@ -190,8 +209,8 @@ def edit(path):
     """
     # Extract the YAML portion of the original file:
     with generic_io.get_file(path, mode="r") as fd:
-        if util.get_file_type(fd) != util.FileType.ASDF:
-            print(f"Error: '{path}' is not an ASDF file.")  # noqa: T201
+        if fd.peek(len(constants.ASDF_MAGIC)) != constants.ASDF_MAGIC:
+            print(f"Error: '{path}' is not an ASDF file.")
             return 1
 
         original_content, available_bytes, contains_blocks = read_yaml(fd)
@@ -217,14 +236,14 @@ def edit(path):
                 new_content = f.read()
 
             if new_content == original_content:
-                print("No changes made to file")  # noqa: T201
+                print("No changes made to file")
                 return 0
 
             try:
                 new_asdf_version = parse_asdf_version(new_content)
                 new_yaml_version = parse_yaml_version(new_content)
-            except Exception as e:  # noqa: BLE001
-                print(f"Error: failed to parse ASDF header: {str(e)}")  # noqa: T201
+            except Exception as e:
+                print(f"Error: failed to parse ASDF header: {e!s}")
                 choice = request_input("(c)ontinue editing or (a)bort? ", ["c", "a"])
                 if choice == "a":
                     return 1
@@ -232,7 +251,7 @@ def edit(path):
                 continue
 
             if new_asdf_version != original_asdf_version or new_yaml_version != original_yaml_version:
-                print("Error: cannot modify ASDF Standard or YAML version using this tool.")  # noqa: T201
+                print("Error: cannot modify ASDF Standard or YAML version using this tool.")
                 choice = request_input("(c)ontinue editing or (a)bort? ", ["c", "a"])
                 if choice == "a":
                     return 1
@@ -240,13 +259,17 @@ def edit(path):
                 continue
 
             try:
-                # Blocks are not read during validation, so this will not raise
-                # an error even though we're only opening the YAML portion of
-                # the file.
-                with open_asdf(io.BytesIO(new_content), _force_raw_types=True):
-                    pass
+                # check this is an ASDF file
+                if new_content[: len(constants.ASDF_MAGIC)] != constants.ASDF_MAGIC:
+                    msg = "Does not appear to be a ASDF file."
+                    raise ValueError(msg)
+                # read the tagged tree (which also checks if the YAML is valid)
+                tagged_tree = util.load_yaml(io.BytesIO(new_content), tagged=True)
+                # validate the tagged tree
+                ctx = AsdfFile(version=new_asdf_version)
+                schema.validate(tagged_tree, ctx=ctx, reading=True)
             except yaml.YAMLError as e:
-                print("Error: failed to parse updated YAML:")  # noqa: T201
+                print("Error: failed to parse updated YAML:")
                 print_exception(e)
                 choice = request_input("(c)ontinue editing or (a)bort? ", ["c", "a"])
                 if choice == "a":
@@ -255,7 +278,7 @@ def edit(path):
                 continue
 
             except schema.ValidationError as e:
-                print("Warning: updated ASDF tree failed validation:")  # noqa: T201
+                print("Warning: updated ASDF tree failed validation:")
                 print_exception(e)
                 choice = request_input("(c)ontinue editing, (f)orce update, or (a)bort? ", ["c", "f", "a"])
                 if choice == "a":
@@ -264,8 +287,8 @@ def edit(path):
                 if choice == "c":
                     continue
 
-            except Exception as e:  # noqa: BLE001
-                print("Error: failed to read updated file as ASDF:")  # noqa: T201
+            except Exception as e:
+                print("Error: failed to read updated file as ASDF:")
                 print_exception(e)
                 choice = request_input("(c)ontinue editing or (a)bort? ", ["c", "a"])
                 if choice == "a":
@@ -291,7 +314,7 @@ def edit(path):
     else:
         # File does not have sufficient space, and binary blocks
         # are present.
-        print("Warning: updated YAML larger than allocated space.  File must be rewritten.")  # noqa: T201
+        print("Warning: updated YAML larger than allocated space.  File must be rewritten.")
         choice = request_input("(c)ontinue or (a)bort? ", ["c", "a"])
         if choice == "a":
             return 1
@@ -346,7 +369,7 @@ def print_exception(e):
     if len(lines) > 20:
         lines = lines[0:20] + ["..."]
     for line in lines:
-        print(f"    {line}")  # noqa: T201
+        print(f"    {line}")
 
 
 def request_input(message, choices):
@@ -366,7 +389,7 @@ def request_input(message, choices):
         if choice in choices:
             return choice
 
-        print(f"Invalid choice: {choice}")  # noqa: T201
+        print(f"Invalid choice: {choice}")
 
     return None
 
@@ -378,4 +401,4 @@ def open_editor(path):
     editor = os.environ.get("EDITOR", DEFAULT_EDITOR)
     # Marked safe because the editor command is specified by an
     # environment variable that the user controls.
-    subprocess.run(f"{editor} {path}", check=True, shell=True)  # nosec
+    subprocess.run(f"{editor} {path}", check=True, shell=True)  # noqa: S602
