@@ -1,22 +1,20 @@
-from asdf import _types
-
-from .complex import ComplexType
-from .constant import ConstantType
+from .constant import Constant
 from .external_reference import ExternalArrayReference
 from .integer import IntegerType
 from .ndarray import NDArrayType
+from .stream import Stream
 
 __all__ = [
     "AsdfObject",
+    "Constant",
     "Software",
     "HistoryEntry",
     "ExtensionMetadata",
     "SubclassMetadata",
-    "ConstantType",
     "NDArrayType",
-    "ComplexType",
     "IntegerType",
     "ExternalArrayReference",
+    "Stream",
 ]
 
 
@@ -24,35 +22,15 @@ class AsdfObject(dict):
     pass
 
 
-class AsdfObjectType(_types.AsdfType):
-    name = "core/asdf"
-    version = "1.1.0"
-    supported_versions = {"1.0.0", "1.1.0"}
-    types = [AsdfObject]
-
-    @classmethod
-    def from_tree(cls, node, ctx):
-        return AsdfObject(node)
-
-    @classmethod
-    def to_tree(cls, data, ctx):
-        return dict(data)
+class Software(dict):
+    pass
 
 
-class Software(dict, _types.AsdfType):
-    name = "core/software"
-    version = "1.0.0"
+class HistoryEntry(dict):
+    pass
 
 
-class HistoryEntry(dict, _types.AsdfType):
-    name = "core/history_entry"
-    version = "1.0.0"
-
-
-class ExtensionMetadata(dict, _types.AsdfType):
-    name = "core/extension_metadata"
-    version = "1.0.0"
-
+class ExtensionMetadata(dict):
     @property
     def extension_uri(self):
         return self.get("extension_uri")
@@ -66,7 +44,7 @@ class ExtensionMetadata(dict, _types.AsdfType):
         return self.get("software")
 
 
-class SubclassMetadata(dict, _types.AsdfType):
+class SubclassMetadata(dict):
     """
     The tagged object supported by this class is part of
     an experimental feature that has since been dropped
@@ -74,6 +52,3 @@ class SubclassMetadata(dict, _types.AsdfType):
     files that used that feature will still deserialize
     without warnings.
     """
-
-    name = "core/subclass_metadata"
-    version = "1.0.0"
