@@ -123,11 +123,12 @@ class _TreeModificationContext:
         duration of the context.
         """
         if id(node) in self._pending:
-            raise RuntimeError(
+            msg = (
                 "Unhandled cycle in tree.  This is possibly a bug "
                 "in extension code, which should be yielding "
                 "nodes that may contain reference cycles."
             )
+            raise RuntimeError(msg)
 
         self._pending.add(id(node))
         try:
@@ -185,7 +186,8 @@ class _TreeModificationContext:
             # modified node is being replaced, which is an
             # error because it breaks references within the
             # tree.
-            raise RuntimeError("Node already has an associated result")
+            msg = "Node already has an associated result"
+            raise RuntimeError(msg)
 
         self._map[id(node)] = (node, result)
 
@@ -265,7 +267,8 @@ def walk_and_modify(top, callback, ignore_implicit_conversion=False, postorder=T
     """
     callback_arity = callback.__code__.co_argcount
     if callback_arity < 1 or callback_arity > 2:
-        raise ValueError("Expected callback to accept one or two arguments")
+        msg = "Expected callback to accept one or two arguments"
+        raise ValueError(msg)
 
     def _handle_generator(result):
         # If the result is a generator, generate one value to
@@ -279,10 +282,7 @@ def walk_and_modify(top, callback, ignore_implicit_conversion=False, postorder=T
         return result
 
     def _handle_callback(node, json_id):
-        if callback_arity == 1:
-            result = callback(node)
-        else:
-            result = callback(node, json_id)
+        result = callback(node) if callback_arity == 1 else callback(node, json_id)
 
         return _handle_generator(result)
 
@@ -300,10 +300,9 @@ def walk_and_modify(top, callback, ignore_implicit_conversion=False, postorder=T
                 # need to fill in the real value later.
                 pending_items[key] = value
                 result[key] = PendingValue
-            else:
-                value = _recurse(value, json_id)
-                if value is not RemoveNode:
-                    result[key] = value
+
+            elif (val := _recurse(value, json_id)) is not RemoveNode:
+                result[key] = val
 
         yield result
 
@@ -311,9 +310,8 @@ def walk_and_modify(top, callback, ignore_implicit_conversion=False, postorder=T
             # Now that we've yielded, the pending children should
             # be available.
             for key, value in pending_items.items():
-                value = _recurse(value, json_id)
-                if value is not RemoveNode:
-                    result[key] = value
+                if (val := _recurse(value, json_id)) is not RemoveNode:
+                    result[key] = val
                 else:
                     # The callback may have decided to delete
                     # this node after all.
@@ -441,10 +439,11 @@ def get_children(node):
     """
     if isinstance(node, dict):
         return list(node.items())
-    elif isinstance(node, list) or isinstance(node, tuple):
+
+    if isinstance(node, (list, tuple)):
         return list(enumerate(node))
-    else:
-        return []
+
+    return []
 
 
 def is_container(node):
@@ -462,4 +461,4 @@ def is_container(node):
     bool
         True if node is a container, False otherwise
     """
-    return isinstance(node, dict) or isinstance(node, list) or isinstance(node, tuple)
+    return isinstance(node, (dict, list, tuple))
