@@ -1,5 +1,3 @@
-# noqa: INP001
-
 import json
 import os
 import subprocess
@@ -14,6 +12,7 @@ from common import assert_file_correct, generate_file
 from packaging.version import Version
 
 import asdf
+from asdf.versioning import AsdfVersion
 
 # Strange version present on pypi that doesn't parse as a Version
 BAD_VERSIONS = {"0"}
@@ -26,7 +25,13 @@ MIN_VERSION_NEW_FILES = Version("2.7.0")
 # Minimum library version to produce files read by the current
 # version of the code.  Earlier versions aren't able to generate
 # files for all the ASDF Standard versions that they claim to support.
-MIN_VERSION_OLD_FILES = Version("2.3.0")
+MIN_VERSION_OLD_FILES = Version("2.5.0")
+
+# Minimum library version that is capable of correctly writing
+# files that follow ASDF Standard 1.0.0.
+# ASDF library versions older than this version would sometimes write newer
+# tags that were not listed in the supported standard
+MIN_VERSION_1_0_0_STANDARD = Version("2.7.0")
 
 GENERATE_SCRIPT_PATH = Path(__file__).parent / "generate_file.py"
 ASSERT_SCRIPT_PATH = Path(__file__).parent / "assert_file_correct.py"
@@ -77,7 +82,7 @@ def fetch_latest_patch_versions(package_name):
 # Enable internet here, otherwise pytest_remotedata will complain
 # (and @pytest.mark.remote_data doesn't work on non-test methods).
 with internet_temporarily_enabled():
-    PATCH_VERSIONS = fetch_latest_patch_versions("asdf")
+    PATCH_VERSIONS = sorted(fetch_latest_patch_versions("asdf"))
 
 
 def env_run(env_path, command, *args, **kwargs):
@@ -85,7 +90,7 @@ def env_run(env_path, command, *args, **kwargs):
     Run a command on the context of the virtual environment at
     the specified path.
     """
-    return subprocess.run([env_path / "bin" / command, *list(args)], **kwargs).returncode == 0
+    return subprocess.run([env_path / "bin" / command, *list(args)], **kwargs).returncode == 0  # noqa: S603
 
 
 def env_check_output(env_path, command, *args):
@@ -93,7 +98,7 @@ def env_check_output(env_path, command, *args):
     Run a command on the context of the virtual environment at
     the specified path, and return the output.
     """
-    return subprocess.check_output([env_path / "bin" / command, *list(args)]).decode("utf-8").strip()
+    return subprocess.check_output([env_path / "bin" / command, *list(args)]).decode("utf-8").strip()  # noqa: S603
 
 
 def get_supported_versions(env_path):
@@ -115,7 +120,7 @@ def get_installed_version(env_path):
     return Version(env_check_output(env_path, "python3", "-c", script))
 
 
-@pytest.fixture(scope="module", params=PATCH_VERSIONS)
+@pytest.fixture(scope="module", params=PATCH_VERSIONS, ids=[f"asdf=={v}" for v in PATCH_VERSIONS])
 def asdf_version(request):
     """
     The (old) version of the asdf library under test.
@@ -130,7 +135,7 @@ def env_path(asdf_version, tmp_path_factory):
     """
     path = tmp_path_factory.mktemp(f"asdf-{asdf_version}-env", numbered=False)
 
-    assert subprocess.run(["virtualenv", str(path)]).returncode == 0
+    assert subprocess.run(["virtualenv", str(path)]).returncode == 0  # noqa: S603,S607
 
     assert env_run(
         path,
@@ -144,7 +149,7 @@ def env_path(asdf_version, tmp_path_factory):
 
 
 @pytest.fixture(autouse=True)
-def _pushd_tmpdir(tmpdir):
+def _pushd_tmp_path(tmp_path):
     """
     Change the working directory, in case the user is running these
     tests from the repo root.  Python will import a module from the
@@ -152,13 +157,13 @@ def _pushd_tmpdir(tmpdir):
     from accidentally comparing the current library code to itself.
     """
     original_cwd = os.getcwd()
-    tmpdir.chdir()
+    os.chdir(str(tmp_path))
     yield
     os.chdir(original_cwd)
 
 
 @pytest.mark.remote_data()
-def test_file_compatibility(asdf_version, env_path, tmpdir):
+def test_file_compatibility(asdf_version, env_path, tmp_path):
     # Sanity check to ensure we're not accidentally comparing
     # the current code to itself.
     installed_version = get_installed_version(env_path)
@@ -173,14 +178,17 @@ def test_file_compatibility(asdf_version, env_path, tmpdir):
     old_supported_versions = set(get_supported_versions(env_path))
     standard_versions = list(current_supported_versions.intersection(old_supported_versions))
 
+    if asdf_version < MIN_VERSION_1_0_0_STANDARD:
+        standard_versions.remove(AsdfVersion("1.0.0"))
+
     # Confirm that this test isn't giving us a false sense of security.
     assert len(standard_versions) > 0
 
-    for standard_version in standard_versions:
+    for standard_version in sorted(standard_versions):
         # Confirm that a file generated by the current version of the code
         # can be read by the older version of the library.
         if asdf_version >= MIN_VERSION_NEW_FILES:
-            current_file_path = Path(str(tmpdir)) / "test-current.asdf"
+            current_file_path = Path(str(tmp_path)) / "test-current.asdf"
             generate_file(current_file_path, standard_version)
             assert env_run(env_path, "python3", ASSERT_SCRIPT_PATH, current_file_path, capture_output=True), (
                 f"asdf library version {asdf_version} failed to read an ASDF Standard {standard_version} "
@@ -190,7 +198,7 @@ def test_file_compatibility(asdf_version, env_path, tmpdir):
         # Confirm that a file generated by the older version of the library
         # can be read by the current version of the code.
         if asdf_version >= MIN_VERSION_OLD_FILES:
-            old_file_path = Path(str(tmpdir)) / "test-old.asdf"
+            old_file_path = Path(str(tmp_path)) / "test-old.asdf"
             assert env_run(
                 env_path,
                 "python3",
@@ -198,10 +206,7 @@ def test_file_compatibility(asdf_version, env_path, tmpdir):
                 old_file_path,
                 str(standard_version),
                 capture_output=True,
-            ), "asdf library version {} failed to generate an ASDF Standard {} file".format(
-                asdf_version,
-                standard_version,
-            )
+            ), f"asdf library version {asdf_version} failed to generate an ASDF Standard {standard_version} file"
             assert_file_correct(old_file_path), (
                 f"asdf library version {asdf_version} produced an ASDF Standard {standard_version}"
                 "that this code failed to read"
