@@ -1,7 +1,6 @@
 import io
 import os
 import pathlib
-import warnings
 from dataclasses import dataclass
 
 import numpy as np
@@ -65,15 +64,8 @@ class AsdfSchemaFile(pytest.File):
         xfail_tests=None,
         **kwargs,
     ):
-        # Fix for depreciation of fspath in pytest 7+
-        from asdf.util import minversion
-
-        if minversion("pytest", "7.0.0"):
-            path = pathlib.Path(fspath)
-            kwargs["path"] = path
-        else:
-            path = fspath
-            kwargs["fspath"] = path
+        path = pathlib.Path(fspath)
+        kwargs["path"] = path
 
         if hasattr(super(), "from_parent"):
             result = super().from_parent(parent, **kwargs)
@@ -218,9 +210,8 @@ class AsdfSchemaExampleItem(pytest.Item):
         return result
 
     def runtest(self):
-        from asdf import AsdfFile, block, util
+        from asdf import AsdfFile, _block, generic_io, util
         from asdf._tests import _helpers as helpers
-        from asdf.exceptions import AsdfDeprecationWarning
 
         # Make sure that the examples in the schema files (and thus the
         # ASDF standard document) are valid.
@@ -239,22 +230,16 @@ class AsdfSchemaExampleItem(pytest.Item):
             util.filepath_to_url(os.path.abspath(os.path.join(os.path.dirname(self.filename), "external.asdf")))
         ] = ff2
 
-        # Add some dummy blocks so that the ndarray examples work
-        for _ in range(3):
-            b = block.Block(np.zeros((1024 * 1024 * 8), dtype=np.uint8))
-            b._used = True
-            ff._blocks.add(b)
-        b._array_storage = "streamed"
+        wb = _block.writer.WriteBlock(np.zeros(1024 * 1024 * 8, dtype=np.uint8))
+        with generic_io.get_file(buff, mode="rw") as f:
+            f.seek(0, 2)
+            _block.writer.write_blocks(f, [wb, wb], streamed_block=wb)
+            f.seek(0)
 
         try:
-            # Do not tolerate any warnings that occur during schema validation
-            with warnings.catch_warnings():
-                warnings.simplefilter("error")
-                warnings.simplefilter("default", category=AsdfDeprecationWarning)
-
-                ff._open_impl(ff, buff, mode="rw")
+            ff._open_impl(ff, buff, mode="rw")
         except Exception:
-            print(f"Example: {self.example.description}\n From file: {self.filename}")  # noqa: T201
+            print(f"Example: {self.example.description}\n From file: {self.filename}")
             raise
 
         # Just test we can write it out.  A roundtrip test
@@ -290,7 +275,7 @@ def _parse_test_list(content):
     return result
 
 
-def pytest_collect_file(path, parent):
+def pytest_collect_file(file_path, parent):
     if not (parent.config.getini("asdf_schema_tests_enabled") or parent.config.getoption("asdf_tests")):
         return None
 
@@ -307,14 +292,14 @@ def pytest_collect_file(path, parent):
     skip_tests = _parse_test_list(parent.config.getini("asdf_schema_skip_tests"))
     xfail_tests = _parse_test_list(parent.config.getini("asdf_schema_xfail_tests"))
 
-    schema_roots = [os.path.join(str(parent.config.rootdir), os.path.normpath(root)) for root in schema_roots]
+    schema_roots = [os.path.join(str(parent.config.rootpath), os.path.normpath(root)) for root in schema_roots]
 
-    if path.ext != ".yaml":
+    if file_path.suffix != ".yaml":
         return None
 
     for root in schema_roots:
-        if str(path).startswith(root) and path.purebasename not in skip_names:
-            posix_path = pathlib.Path(path).as_posix()
+        if str(file_path).startswith(root) and file_path.stem not in skip_names:
+            posix_path = pathlib.Path(file_path).as_posix()
             schema_skip_tests = []
             for suffix, names in skip_tests.items():
                 if posix_path.endswith(suffix):
@@ -326,8 +311,8 @@ def pytest_collect_file(path, parent):
 
             return AsdfSchemaFile.from_parent(
                 parent,
-                fspath=path,
-                skip_examples=(path.purebasename in skip_examples),
+                fspath=file_path,
+                skip_examples=(file_path.stem in skip_examples),
                 validate_default=validate_default,
                 ignore_unrecognized_tag=ignore_unrecognized_tag,
                 ignore_version_mismatch=ignore_version_mismatch,
